import glob
import click
from tabulate import tabulate

from qiskitflow.utils.constants import EXPERIMENTS_DIRECTORY, PAGINATION_SIZE
from qiskitflow.lib.experiment import Experiment


def get_experiments():
    experiments = set()
    for path in glob.glob("{}/**/**/run.json".format(EXPERIMENTS_DIRECTORY)):
        experiment_name, _, _ = path.split("/")[-3:]
        experiments.add(experiment_name)
    experiments = [[e] for e in experiments]

    for i in range(0, len(experiments), PAGINATION_SIZE):
        if i != 0:
            click.confirm("Next page?", abort=True)
        table = experiments[i: i + PAGINATION_SIZE]
        click.echo(click.style(tabulate(table, ["Experiment"], tablefmt='grid'), fg='white'))


import unittest
from click.testing import CliRunner

from qiskitflow.cli.cli import run_detailed
from qiskitflow.lib.experiment import Experiment


class TestCLIExperimentInfo(unittest.TestCase):
    """ Tests CLI experiment info. """

    def setUp(self) -> None:
        self.runner = CliRunner()

    def test_experiment_info(self):
        """ Test experiments list. """
        with self.runner.isolated_filesystem():
            with Experiment("experiment") as exp:
                exp.run_id = "run_for_testing"
                exp.write_metric("metric", 1.)
                exp.write_metric("other metric", 2.)
                exp.write_parameter("parameter", "value")
                exp.write_parameter("other parameter", "other value")

            result = self.runner.invoke(run_detailed, ["run_for_testing"])

            self.assertEqual(result.exit_code, 0)
            for metric in exp.metrics:
                self.assertTrue(str(metric) in result.output)
            for param in exp.parameters:
                self.assertTrue(str(param) in result.output)




import unittest
from click.testing import CliRunner

from qiskitflow.cli.cli import runs_list
from qiskitflow.lib.experiment import Experiment


class TestCLIExperimentList(unittest.TestCase):
    """ Tests CLI experiments list. """

    def setUp(self) -> None:
        self.runner = CliRunner()

    def test_experiments_list(self):
        """ Test experiments list. """
        with self.runner.isolated_filesystem():
            for idx in range(5):
                with Experiment("{} experiment".format(idx)) as exp:
                    exp.write_metric("metric", idx * 1.)
                    exp.write_parameter("parameter", "value {}".format(idx))
            result = self.runner.invoke(runs_list)

            self.assertEqual(result.exit_code, 0)
            for idx in range(5):
                self.assertTrue("{} experiment".format(idx) in result.output)




from typing import Union

import time


class Metric:
    def __init__(self, name: str, value: Union[float, int]):
        """ Experiment metric.

        Args:
            name (str): name of metric
            value (float|int): value of metric
        """

        self.name = name
        self.value = value
        self.timestamp = int(time.time())

    def __dict__(self):
        return {
            "name": self.name,
            "value": self.value,
            "timestamp": self.timestamp
        }

    def __eq__(self, other: 'Metric'):
        return self.name == other.name and self.value == other.value

    def __repr__(self):
        return "{}:{}".format(self.name, self.value)





import glob
import click
from tabulate import tabulate
from datetime import datetime

from qiskitflow.utils.constants import EXPERIMENTS_DIRECTORY, PAGINATION_SIZE
from qiskitflow.lib.experiment import Experiment


def detail(run_id: str) -> None:
    click.echo(click.style("Run [{}] detailed information.\n".format(run_id), fg='magenta'))

    glob_query = "{}/**/{}/run.json".format(EXPERIMENTS_DIRECTORY, run_id)
    results = glob.glob(glob_query)

    if len(results) == 0:
        click.echo(click.style("No data for experiment found =(\n"
                               "Check `qiskitflow runs`", fg="yellow"))

    for path in results:
        run = Experiment.load(path)

        ps = "\n".join(str(p) for p in run.parameters)
        ms = "\n".join(str(m) for m in run.metrics)
        cts = "\n".join(str(c) for c in run.counts)

        header = ["", "Values"]
        rows = [
            ["Run ID", run.run_id],
            ["Experiment", run.name],
            ["Date", datetime.fromtimestamp(run.timestamp)],
            ["Parameters", ps],
            ["Metrics", ms],
            ["Counts", cts]
        ]

        click.echo(click.style(tabulate(rows, header, tablefmt='grid'), fg='white'))



import glob
import click
import os
from tabulate import tabulate
from datetime import datetime

from qiskitflow.utils.constants import EXPERIMENTS_DIRECTORY, PAGINATION_SIZE
from qiskitflow.lib.experiment import Experiment


def _get_runs(search: str,
              experiment_name: str,
              order_by: str,
              order_type: str = "asc"):

    glob_query = "{}/**/**/run.json".format(EXPERIMENTS_DIRECTORY)
    if experiment_name:
        glob_query = "{}/{}/**/run.json".format(EXPERIMENTS_DIRECTORY, experiment_name)

    runs_paths = glob.glob(glob_query)

    # gather all runs and sort them accordingly
    runs = []
    for path in runs_paths:
        experiment = Experiment.load(path)
        if experiment_name and experiment_name != experiment.name:
            continue
        if search in experiment.run_id:
            ordering = experiment.get_metric(order_by) if order_by else os.path.getctime(path)
            runs.append({
                "order": int(ordering),
                "run": Experiment.load(path)
            })
    runs.sort(key=lambda x: x["order"], reverse=order_type == "desc")

    # collect rows to display
    rows = []
    for entry in runs:
        run = entry["run"]
        ps = [str(p) for p in run.parameters]
        ms = [str(m) for m in run.metrics]

        rows.append([datetime.fromtimestamp(run.timestamp),
                     "{}\n({})".format(run.name, run.run_id),
                     "\n".join(ms),
                     "\n".join(ps)])

    return rows


def runs(search: str,
         experiment: str,
         order_by: str,
         order_type: str = "asc") -> None:
    """ Show list of executed experiment runs based on parameters.

    Args:
        search (str): Search by run name
        experiment (str): Filter by experiment name
        order_by (str): Order by metric
        order_type (str): Order type
    """
    click.echo(click.style("Experiment runs\n", fg='magenta'))
    runs = _get_runs(search=search.strip(),
                     experiment_name=experiment.strip(),
                     order_by=order_by.strip(),
                     order_type=order_type.strip())
    headers = ["Date", "Name", "Metrics", "Parameters"]

    for i in range(0, len(runs), PAGINATION_SIZE):
        if i != 0:
            click.confirm("Next page?", abort=True)
        table = runs[i: i + PAGINATION_SIZE]
        click.echo(click.style(tabulate(table, headers, tablefmt='grid'), fg='white'))




import sys
import uuid
import os
import math
import json
import shutil
import time

from typing import Optional, Union
from qiskitflow.utils.constants import EXPERIMENTS_DIRECTORY
from qiskitflow.lib.models import Metric, Parameter, Count, Artifact
from qiskitflow.utils.utils import include_patterns
from qiskitflow._version import __version__


class Experiment:
    def __init__(self,
                 name: str, 
                 entrypoint: Optional[str] = None,
                 sourcecode_dir: Optional[str] = None,
                 save_path: Optional[str] = None):
        """ Experiment.
        
        Args:
            name (str): name of experiment
            entrypoint (str): script that were used to run this experiment
            sourcecode_dir (str): path to directory with sourcecode and entrypoint
            save_path (str): experiments save location
        """
        if not save_path:
            save_path = "./"
        self.save_path = save_path
        self.entrypoint = entrypoint
        self.sourcecode_dir = sourcecode_dir

        self.name = name
        self.run_id = str(uuid.uuid4().hex)

        self.metrics = []
        self.parameters = []
        self.counts = []
        self.state_vectors = []  # TODO: implement

        self.artifacts = set()

        self.timestamp = int(time.time())

        self.version = __version__

        self.timestamp = int(time.time())

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._save_experiment()

    def write_metric(self, metric_name: str, metric_value: Union[float, int]):
        """ Writes metric to experiment run. """
        self.metrics.append(Metric(metric_name, metric_value))

    def write_parameter(self, parameter_name: str, parameter_value: Union[str, float, int]):
        """ Writes parameter to experiment run. """
        self.parameters.append(Parameter(parameter_name, parameter_value))

    def write_counts(self, name: str, counts: dict):
        """ Writes meas counts to experiment run. """
        self.counts.append(Count(name, counts))

    def write_image(self, name: str, image_path: str):
        """ Writes image to experiment run. """
        self.write_artifact(name, image_path)

    def write_artifact(self, name: str, path: str):
        """ Write artifact to experiment run. """
        self.artifacts.add(Artifact(name, path))

    def set_run(self, run_id: str):
        """ Set run id for experiment. """
        self.run_id = run_id

    def _save_experiment(self):
        """ Saves experiment run. """
        run_dir, sourcecode_directory = self._create_and_get_save_directory()

        # saving files
        if self.entrypoint and self.sourcecode_dir and os.path.isdir(self.sourcecode_dir):
            shutil.copytree(self.sourcecode_dir, sourcecode_directory,
                            ignore=include_patterns("*.py", "*.ipynb", "requirements.txt"))

        for art in self.artifacts:
            art.save(run_dir)

        with open("{}/run.json".format(run_dir), "w") as f:
            json.dump(self.__dict__(), f)
        return self.run_id

    @classmethod
    def load(cls, path: str):
        """ Load experiment run from file. """
        with open(path, "r") as f:
            run_data = json.load(f)
            exp = Experiment(run_data["name"], entrypoint=run_data["entrypoint"])

            version = run_data.get("version")
            exp.version = version

            run_id = run_data.get("run_id")
            exp.run_id = run_id

            metrics = []
            for m in run_data["metrics"]:
                metrics.append(Metric(m["name"], m["value"]))

            parameters = []
            for p in run_data["parameters"]:
                parameters.append(Parameter(p["name"], p["value"]))

            counts = []
            for cnt in run_data["counts"]:
                counts.append(Count(cnt["name"],
                                    cnt["value"]))

            artifacts = set()
            for art in run_data.get("artifacts", []):
                artifacts.add(Artifact(art.get("name"), art.get("path")))

            exp.metrics = metrics
            exp.parameters = parameters
            exp.counts = counts
            exp.artifacts = artifacts
            exp.timestamp = run_data.get("timestamp", exp.timestamp)

            # TODO: state vector
            return exp

    def _create_and_get_save_directory(self) -> [str, str]:
        """ Creates directory for experiment run if not exists and return path. """
        directory = "{}/{}/{}/{}".format(self.save_path, EXPERIMENTS_DIRECTORY, self.name, self.run_id)
        sourcecode_directory = "{}/sourcecode".format(directory)
        if not os.path.exists(directory):
            os.makedirs(directory)
        else:
            raise Exception("Experiment run [{}] already exists for experiment [{}]".format(self.run_id, self.name))
        return directory, sourcecode_directory

    def __dict__(self):
        return {
            "version": self.version,
            "name": self.name,
            "run_id": self.run_id,
            "metrics": [m.__dict__() for m in self.metrics],
            "parameters": [p.__dict__() for p in self.parameters],
            "counts": [c.__dict__() for c in self.counts],
            "artifacts": [a.__dict__() for a in self.artifacts],
            "entrypoint": self.entrypoint,
            "timestamp": self.timestamp
        }

    def __repr__(self):
        return 'Experiment {} (run: {})'.format(self.name, self.run_id)

    def get_metric(self, name: str) -> Union[float, int]:
        """ Returns metric value. """
        metrics = {m.name: m.value for m in self.metrics}
        return metrics.get(name, sys.maxsize)



import os
import unittest
import shutil

from qiskitflow import Experiment
from qiskitflow.utils.constants import EXPERIMENTS_DIRECTORY


class TestExperiment(unittest.TestCase):
    """ Experiment tests. """

    def setUp(self):
        self.resources_dir = "{}/../resources/".format(os.path.dirname(os.path.abspath(__file__)))

    def tearDown(self) -> None:
        test_experiments_location = "{}/{}".format(self.resources_dir, EXPERIMENTS_DIRECTORY)
        if os.path.exists(test_experiments_location):
            shutil.rmtree(test_experiments_location)

    def test_experiment_saving_and_loading(self):
        """ Base test for experiment save. """
        with Experiment("test experiment",
                        save_path=self.resources_dir) as exp:
            exp.write_metric("test metric", 0.1)
            exp.write_metric("test metric 2", 2)

            exp.write_parameter("test parameter", "test parameter value")
            exp.write_parameter("test parameter 2", "test paraeter value 2")

            exp.write_counts("counts", {"00": 1024, "11": 0})

            run_id = exp.run_id

        restored = Experiment.load("{}/{}/test experiment/{}/run.json".format(self.resources_dir,
                                                                              EXPERIMENTS_DIRECTORY,
                                                                              run_id))

        self.assertEqual(restored.name, exp.name)

        self.assertEqual(len(restored.metrics), len(exp.metrics))
        self.assertEqual(len(restored.parameters), len(exp.parameters))
        self.assertEqual(len(restored.counts), len(exp.counts))

        self.assertEqual(restored.metrics, exp.metrics)
        self.assertEqual(restored.parameters, exp.parameters)
        self.assertEqual(restored.counts, exp.counts)

    def test_artifact_save(self):
        """ Tests saving artifacts. """
        run_id = None
        with Experiment("artifact saving experiment", save_path=self.resources_dir) as exp:
            exp.write_artifact("artifact", "{}/artifact.txt".format(self.resources_dir))
            run_id = exp.run_id

        self.assertTrue(os.path.isfile("{}/../resources/{}/artifact saving experiment/{}/"
                                       "artifact.txt".format(self.resources_dir,
                                                             EXPERIMENTS_DIRECTORY,
                                                             run_id)))

    def test_sourcecode_save(self):
        """ Tests saving of sourecode files. """
        run_id = None

        with Experiment("test experiment",
                        entrypoint="entrypoint_example.py",
                        save_path=self.resources_dir,
                        sourcecode_dir=self.resources_dir) as exp:
            run_id = exp.run_id

        self.assertTrue(os.path.isfile("{}/../resources/{}/test experiment/{}/sourcecode/"
                                       "entrypoint_example.py".format(self.resources_dir,
                                                                      EXPERIMENTS_DIRECTORY,
                                                                      run_id)))
        # should skip non pythonic files
        self.assertFalse(os.path.isfile("{}/../resources/{}/test experiment/{}/sourcecode/"
                                        "otherfile.txt".format(self.resources_dir,
                                                               EXPERIMENTS_DIRECTORY,
                                                               run_id)))



import fnmatch
from os.path import isdir, join

from .constants import EXPERIMENTS_DIRECTORY


def get_experiment_folder(name: str) -> str:
    return "{}/{}".format(EXPERIMENTS_DIRECTORY, name)


def include_patterns(*patterns):
    """ Function that can be used as shutil.copytree() ignore parameter that
    determines which files *not* to ignore, the inverse of "normal" usage.

    This is a factory function that creates a function which can be used as a
    callable for copytree()'s ignore argument, *not* ignoring files that match
    any of the glob-style patterns provided.

    ‛patterns’ are a sequence of pattern strings used to identify the files to
    include when copying the directory tree.

    Example usage:

        copytree(src_directory, dst_directory,
                 ignore=include_patterns('*.sldasm', '*.sldprt'))
    """
    def _ignore_patterns(path, all_names):
        # Determine names which match one or more patterns (that shouldn't be
        # ignored).
        keep = (name for pattern in patterns
                        for name in fnmatch.filter(all_names, pattern))
        # Ignore file names which *didn't* match any of the patterns given that
        # aren't directory names.
        dir_names = (name for name in all_names if isdir(join(path, name)) and name == EXPERIMENTS_DIRECTORY)
        return set(all_names) - set(keep) - set(dir_names)

    return _ignore_patterns



from typing import Dict

import time


class Count:
    def __init__(self, name: str, value: Dict[str, int]):
        """ Experiment measurement.

        Args:
            name (str): name of measurement
            value (dict): value of measurement
        """
        self.name = name
        self.value = value
        self.timestamp = int(time.time())

    def __dict__(self):
        return {
            "name": self.name,
            "value": self.value
        }

    def __eq__(self, other: 'Counts'):
        return self.name == other.name and self.value == other.value

    def __repr__(self):
        values = "\n".join("- {} -> {}".format(k, v) for k, v in self.value.items())
        return "{}:\n{}".format(self.name, values)


import time
import shutil
import logging


class Artifact:
    def __init__(self, name: str, path: str):
        """ Artifact for experiment.

        Args:
            name (str): name of artifact
            path (str): path to artifact
        """
        self.name = name
        self.path = path

        self.timestamp = int(time.time())

    def save(self, dst: str):
        """ Save artifact to destination folder. """
        try:
            shutil.copy(self.path, dst)
        except Exception as e:
            logging.warning("Error occurred during saving artifact: {}".format(e))

    def __hash__(self):
        return hash("{}:{}".format(self.name, self.path))

    def __dict__(self):
        return {
            "name": self.name,
            "path": self.path,
            "timestamp": self.timestamp
        }

    def __eq__(self, other):
        return self.name == other.name and self.path == other.path

    def __repr__(self):
        return "Artifact({}: {}...)".format(self.name, self.path[:10])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("qiskitflow/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

setuptools.setup(
    name="qiskitflow",
    version=version,
    author="Iskandar Sitdikov",
    author_email="iskandar.sitdikov@gmail.com",
    description="QiskitFlow. Reproducible quantum experiments.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IceKhan13/QiskitFlow",
    packages=setuptools.find_packages(),
    keywords=['quantum computing', 'qiskit', 'experiments', 'reproducibility'],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'Click>=6',
        'requests>=2.22.0'
    ],
    python_requires='>=3.7',
    entry_points='''
        [console_scripts]
        qiskitflow=qiskitflow.cli.cli:qiskitflow
    '''
)



import click

from qiskitflow.cli.runs.list import runs
from qiskitflow.cli.runs.experiments import get_experiments
from qiskitflow.cli.runs.detail import detail


@click.group(help="""
    Base QiskitFlow CLI function.
""")
def qiskitflow():
    click.echo(click.style("\nQiskitFlow. Reproducible quantum experiments.\n", fg='magenta'))


@qiskitflow.command("runs")
@click.option("--search", default="")
@click.option("--experiment", default="")
@click.option("--order_by", default="")
@click.option("--order_type", default="asc")
def runs_list(search, experiment, order_by, order_type):
    runs(search=search,
         experiment=experiment,
         order_by=order_by,
         order_type=order_type)


@qiskitflow.command("run")
@click.argument("run_id")
def run_detailed(run_id):
    detail(run_id)


@qiskitflow.command("experiments")
def experiments():
    get_experiments()
